unit VaxPipeClient;

interface

uses
  System.Classes, Windows, VaxPipeData,
  System.SysUtils, System.Generics.Collections;


type TVaxPipeClient = class

  private
  m_hHandlePipe: THandle;

  m_hWaitEventRead : THandle;
  m_hWaitEventWrite : THandle;

  m_objWriteDataList: TList<TVaxPipeData>;

  m_hOverlapRead: Overlapped;
  m_hOverlapWrite: Overlapped;

  m_aDataReadPipe : PByte;

  public
  constructor Create();
  destructor Destroy(); override;

  private

  function CreatePipe(sPipeName : String; bEnableWrite : Boolean; var hWaitEventRead: THandle; var hWaitEventWrite: THandle) : Boolean;
  function WritePipe(pData: PByte; nSize: Integer): Boolean;

  public
  function CreatePipeReadWrite(sPipeName : String; var hWaitEventRead: THandle; var hWaitEventWrite: THandle) : Boolean;
  function CreatePipeReadOnly(sPipeName : String; var hWaitEventRead: THandle) : Boolean;

  procedure ClosePipe();

  function BeginReadPipe() : Boolean;
  function EndReadPipe(var aDataRead: PByte; var nSizeRead: Integer) : Boolean;

  procedure BeginWritePipe(pData: PByte; nSize: Integer);
  procedure EndWritePipe();

end;

implementation

  const PIPE_BUFFER_LENGTH = 6400;

  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  constructor TVaxPipeClient.Create();
  begin

    m_hHandlePipe := 0;

    m_hWaitEventRead := 0;
    m_hWaitEventWrite := 0;

    GetMem(m_aDataReadPipe, PIPE_BUFFER_LENGTH);

  end;

  destructor TVaxPipeClient.Destroy();
  begin

    ClosePipe();

    if(m_aDataReadPipe <> nil) Then
    begin
      FreeMem(m_aDataReadPipe);
      m_aDataReadPipe := nil;
    end;

    Inherited;

  end;

  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  function TVaxPipeClient.CreatePipeReadWrite(sPipeName : String; var hWaitEventRead: THandle; var hWaitEventWrite: THandle) : Boolean;
  var
    bResult: Boolean;
  begin

    bResult := CreatePipe(sPipeName, true, hWaitEventRead, hWaitEventWrite);
    Exit(bResult);

  end;

  function TVaxPipeClient.CreatePipeReadOnly(sPipeName : String; var hWaitEventRead: THandle) : Boolean;
  var
    bResult: Boolean;
    hWaitEventWrite : THandle;
  begin

    bResult := CreatePipe(sPipeName, false, hWaitEventRead, hWaitEventWrite);
    Exit(bResult);

  end;

  function TVaxPipeClient.CreatePipe(sPipeName : String; bEnableWrite : Boolean; var hWaitEventRead: THandle; var hWaitEventWrite: THandle) : Boolean;
  var
    dwMode : DWORD;
    dwDesireAccess : DWORD;
  begin

    dwDesireAccess :=  GENERIC_READ OR FILE_WRITE_ATTRIBUTES;

    if (bEnableWrite = True) then
      dwDesireAccess :=  GENERIC_READ OR GENERIC_WRITE OR FILE_WRITE_ATTRIBUTES;

    m_hHandlePipe := CreateFile(PChar(sPipeName), dwDesireAccess, 0, NIL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, 0);

    if (m_hHandlePipe = INVALID_HANDLE_VALUE) then
    begin

      dwMode := GetLastError();

      if (GetLastError() <> ERROR_PIPE_BUSY) then
      begin
        Exit(false);
      end;

      if not WaitNamedPipe(PChar(sPipeName), 3000) then
      begin
        Exit(false);
      end;

    end;

    //Result := (m_hHandlePipe <> INVALID_HANDLE_VALUE);

    dwMode := PIPE_READMODE_MESSAGE;

    SetNamedPipeHandleState(m_hHandlePipe, dwMode, NIL, NIL);

     m_hWaitEventRead := CreateEvent(NIL, False, False, NIL);

     if(m_hWaitEventRead = 0) Then
     begin

      ClosePipe();
      Exit(false);

     end;

     if bEnableWrite then
     begin

      m_hWaitEventWrite := CreateEvent(nil, False, False, nil);

      if m_hWaitEventWrite = 0 then
      begin
        ClosePipe;
        Exit(False);
      end;

      m_objWriteDataList := TList<TVaxPipeData>.Create;

    end;

    hWaitEventRead := m_hWaitEventRead;
    hWaitEventWrite := m_hWaitEventWrite;

    Exit(true);

  end;

  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  procedure TVaxPipeClient.ClosePipe();
  begin

    if Assigned(m_objWriteDataList) then
    begin

      while m_objWriteDataList.Count > 0 do
      begin
        TVaxPipeData(m_objWriteDataList[0]).Free();
        m_objWriteDataList.Delete(0);
      end;

      FreeAndNil(m_objWriteDataList);
    end;

    if(m_hHandlePipe <> 0) Then
    begin
      CloseHandle(m_hHandlePipe);
      m_hHandlePipe := 0;
    end;

    if(m_hWaitEventRead <> 0) Then
    begin
      CloseHandle(m_hWaitEventRead);
      m_hWaitEventRead := 0;
    end;

    if(m_hWaitEventWrite <> 0) Then
    begin
      CloseHandle(m_hWaitEventWrite);
      m_hWaitEventWrite := 0;
    end;

  end;

  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  function TVaxPipeClient.BeginReadPipe() : Boolean;
  var
    bResult: Boolean;
    dwSizeRead: DWORD;
  begin

    FillChar(m_hOverlapRead, sizeof(m_hOverlapRead), #0);
    m_hOverlapRead.hEvent := m_hWaitEventRead;

    dwSizeRead := PIPE_BUFFER_LENGTH;

    bResult := ReadFile(m_hHandlePipe, m_aDataReadPipe^, PIPE_BUFFER_LENGTH, dwSizeRead, @m_hOverlapRead);

    if(bResult = true) Then
    begin
      Exit(true);
    end;

    if ((bResult <> false) OR (GetLastError() <> ERROR_IO_PENDING)) Then
     begin
      Exit(false);
    end;

    //((bResult = false) AND (GetLastError() = ERROR_MORE_DATA)) Then

    Exit(true);

  end;


  function TVaxPipeClient.EndReadPipe(var aDataRead: PByte; var nSizeRead: Integer) : Boolean;
  var
    dwSizeRead: DWORD;
  begin

    if not GetOverlappedResult(m_hHandlePipe, m_hOverlapRead, dwSizeRead, False) Then
      Exit(False);

    nSizeRead := dwSizeRead;
    aDataRead := m_aDataReadPipe;

    Exit(true);

  end;


  //////////////////////////////////////////////////////////////////////////////
  //////////////////////////////////////////////////////////////////////////////

  const
  PIPE_MAX_NO_DATA_MESSAGES = 240;

  procedure TVaxPipeClient.BeginWritePipe(pData: PByte; nSize: Integer);
  var
    objDataWrite: TVaxPipeData;
  begin

    if (m_hHandlePipe = 0) or (m_hWaitEventWrite = 0) then
      Exit;

    if m_objWriteDataList.Count > PIPE_MAX_NO_DATA_MESSAGES then
      Exit;

    objDataWrite := TVaxPipeData.Create(pData, nSize);

    if m_objWriteDataList.count > 0 then
    begin
      m_objWriteDataList.Add(objDataWrite);
      Exit;
    end;

    if (WritePipe(objDataWrite.Data, objDataWrite.Size) = false) then
    begin
      objDataWrite.Free();
      Exit;
    end;

    m_objWriteDataList.Add(objDataWrite);

  end;


  function TVaxPipeClient.WritePipe(pData: PByte; nSize: Integer): Boolean;
  var
    dwBytesWritten: DWORD;
  begin

    if (m_hHandlePipe = 0) or (m_hWaitEventWrite = 0) then
      Exit(False);

    FillChar(m_hOverlapWrite, SizeOf(m_hOverlapWrite), 0);
    m_hOverlapWrite.hEvent := m_hWaitEventWrite;

    if (WriteFile(m_hHandlePipe, pData^, nSize, dwBytesWritten, @m_hOverlapWrite) = false) then
    begin

      var nError := GetLastError();

      if nError = ERROR_IO_PENDING then
        Exit(True);

      if (nError = ERROR_BROKEN_PIPE) or (nError = ERROR_NO_DATA) then
      begin
        // Handle broken pipe or no data case if necessary
      end;

      Exit(False);

    end;

    Exit(True);
  end;


  procedure TVaxPipeClient.EndWritePipe();
  var
    objDataWrite: TVaxPipeData;
  begin

    if (m_hHandlePipe = 0) or (m_hWaitEventWrite = 0) then
      Exit;

    if (m_objWriteDataList.Count = 0) then
      Exit;

    objDataWrite := TVaxPipeData(m_objWriteDataList.First);

    m_objWriteDataList.Delete(0);
    objDataWrite.Free();

    if (m_objWriteDataList.Count = 0) then
      Exit;

    if (WritePipe(objDataWrite.Data, objDataWrite.Size) = False) then
    begin
      Exit;
    end;

  end;


end.
